from kenosis import address
from kenosis import xmlrpclib_transport
from ds import dsqueue
from ds import dsthread
from ds import dsfile
from ds import dstime
from ds import dsunittest
from ds import task
from ds import ui

import copy
from sets import Set
import Queue
import SimpleXMLRPCServer
import os
import socket
import sys
import time
import threading
import urllib2
import xmlrpclib


NetworkErrorClasses = (xmlrpclib.Fault,
                       xmlrpclib.ProtocolError,
                       xmlrpclib.expat.ExpatError,
                       urllib2.URLError,
                       socket.error)
NodeAddressUnknown = address.NodeAddress(stringAddress="1" * address.addressLengthInBits)
socket.setdefaulttimeout(2)

class GrowingQueue:
    def __init__(self, iterator, realQueue):
        self.iterator_ = iterator
        self.realQueue_ = realQueue
        self.fromIterator_ = None
        self.iteratorEmpty_ = 0
        ret = self._getFromIterator()
        assert ret is None

    def empty(self):
        return self.realQueue_.empty() and self.iteratorEmpty_

    def get(self):
        if self.empty():
            raise Queue.Empty()
        assert not self.empty()
        if not self.iteratorEmpty_ and not self.realQueue_.empty():
            fromQueue = self.realQueue_.get()
            if self.fromIterator_ < fromQueue:
                self.realQueue_.put(fromQueue)
                return self._getFromIterator()
            else:
                return fromQueue
        elif self.iteratorEmpty_:
            return self.realQueue_.get()
        else:
            return self._getFromIterator()
                
    def _getFromIterator(self):
        try:
            temp = self.fromIterator_
            self.fromIterator_ = self.iterator_.next()
        except StopIteration:
            self.iteratorEmpty_ = 1
        return temp

def newRpcHeader(sourceNodeAddress, destNodeAddress):
    return {"sourceNodeAddress":sourceNodeAddress,
            "destNodeAddress":destNodeAddress}

def assertRpcHeaderMightBeValid(rpcHeader):
    assert type(rpcHeader) == type({}), "assertRpcHeaderMightBeValid: %s is not a valid rpcHeader" % repr(rpcHeader)
    assert rpcHeader.has_key("sourceNodeAddress")
    assert rpcHeader.has_key("destNodeAddress")

def assertRpcHeaderValid(rpcHeader):
    assertRpcHeaderMightBeValid(rpcHeader=rpcHeader)
    assert rpcHeader.has_key("sourceNetHost")
    assert rpcHeader.has_key("sourceNetPort")

def sourceNodeAddressFrom(rpcHeader):
    return rpcHeader["sourceNodeAddress"]

def sourceNetAddressFrom(rpcHeader):
    return "%s:%s" % (rpcHeader["sourceNetHost"], rpcHeader["sourceNetPort"])

def destNodeAddressFrom(rpcHeader):
    return rpcHeader["destNodeAddress"]

class Rpc:
    def __init__(self, rpcHeader, serverProxy):
        self.rpcHeader_ = rpcHeader
        self.sp_ = serverProxy

    def __getattr__(self, name):
        # Note: we deliberately do not use nameIsPrivate() here
        # becuase that is a stronger test than we need. We only want
        # to prevent python from thinking that we support __eq__, etc.
        if name.startswith("__") and name.endswith("__"):
            raise AttributeError(name)
        attr = getattr(self.sp_, name)
        return Rpc(rpcHeader=self.rpcHeader_, serverProxy=attr)
        
    def __call__(self, *args):
        try:
            return self.sp_(self.rpcHeader_,  *args)
        except NetworkErrorClasses, e:
            raise NetworkError(e)

class NodeInfo:
    version_ = 1
    
    def __init__(self, nodeAddress, netAddress):
        self.nodeAddress_ = nodeAddress
        self.netAddress_ = netAddress
        self.updateFreshness()
    def __repr__(self):
        return "NodeInfo(%s, %s, %s)" % (self.nodeAddress_, self.netAddress_, self.freshness())

    def updateFreshness(self, newNetAddress=None):
        self.created_ = dstime.time()
        if newNetAddress is not None:
            self.netAddress_ = newNetAddress

    def freshness(self):
        return dstime.time() - self.created_

    def netAddress(self):
        return self.netAddress_

    def nodeAddress(self):
        return self.nodeAddress_

    def __eq__(self, otherNodeInfo):
        assert isinstance(otherNodeInfo, NodeInfo)
        return self.nodeAddress_ == otherNodeInfo.nodeAddress_

    def asTuple(self):
        return (self.nodeAddress_, self.netAddress_)

class KenosisError(Exception): pass
class NodeNotFound(KenosisError): pass
class NetworkError(KenosisError): pass
class RpcProtocolError(KenosisError): pass

class _StaleInfo(Exception): pass

def RecursiveConversion(data, matcher, converter):
    try:
        matched = matcher(data)
    except AttributeError:
        matched = 0
    if matched:
        return converter(data)
    elif type(data) == type({}):
        d = {}
        for k, v in data.items():
            k = RecursiveConversion(data=k, matcher=matcher, converter=converter)
            d[k] = RecursiveConversion(data=v, matcher=matcher, converter=converter)
        return d
    elif type(data) in (type(tuple()), type([])):
        d = []
        for v in data:
            d.append(RecursiveConversion(data=v, matcher=matcher, converter=converter))
        return d
    else:
        return data

def ConvertAddressToStringRecursively(data):
    return RecursiveConversion(
        data=data, matcher=lambda x:isinstance(x, address.NodeAddress),
        converter=lambda x:str(x))

def ConvertStringToAddressRecursively(data):
    return RecursiveConversion(
        data=data, matcher=lambda x:x.startswith("address:"),
        converter=lambda x:address.NodeAddress(stringAddress=x[len("address:"):]))

class NodeRpcFilter:
    def __init__(self, nodeKernel):
        self.nodeKernel_ = nodeKernel

    def findNode(self, nodeAddress):
        return self.nodeKernel_.rpcFindNode(nodeAddress=nodeAddress)

    def ping(self):
        return self.nodeKernel_.nodeAddress_

def nameIsPrivate(name):
    return name.startswith("_") or name.endswith("_")

class RpcServerAdapter:
    def __init__(self, adaptee, nodeKernel):
        self.adaptee_ = adaptee
        self.nodeKernel_ = nodeKernel

    def _listPublicMethod(self):
        return SimpleXMLRPCServer.list_public_methods(self.adaptee_)

    def __getattr__(self, name):
        if nameIsPrivate(name=name):
            raise AttributeError(name)
        return RpcServerAdapter(
            adaptee=getattr(self.adaptee_, name),
            nodeKernel=self.nodeKernel_)

    def __call__(self, *args):
        args = ConvertStringToAddressRecursively(args)
        rpcHeader = args[0]
        destNodeAddr = destNodeAddressFrom(rpcHeader=rpcHeader)
        sourceNodeAddr = sourceNodeAddressFrom(rpcHeader=rpcHeader)
        assert destNodeAddr == self.nodeKernel_.nodeAddress_ or \
               destNodeAddr == NodeAddressUnknown, \
               "%s != %s" % (destNodeAddr, self.nodeKernel_.nodeAddress_)

        if sourceNodeAddr != self.nodeKernel_.nodeAddress_:
            self.nodeKernel_._updateRoutingTableWith(
                netAddress=sourceNetAddressFrom(rpcHeader=rpcHeader),
                nodeAddress=sourceNodeAddressFrom(rpcHeader=rpcHeader))

        otherArgs = args[1:]
        assertRpcHeaderValid(rpcHeader=rpcHeader)
        ret = self.adaptee_(*otherArgs)
        return ConvertAddressToStringRecursively(data=ret)

class NodeRpcFrontend:
    def __init__(self, nodeKernel):
        self.nodeKernel_ = nodeKernel
        self.kenosis = RpcServerAdapter(adaptee=NodeRpcFilter(nodeKernel=nodeKernel), nodeKernel=self.nodeKernel_)
        self.kenosis.supportsService = self.supportsService

    def supportsService(self, secondSelf, name):
        if nameIsPrivate(name=name):
            return False
        return hasattr(self, name)
    
    def registerNamedHandler(self, name, handler):
        if nameIsPrivate(name=name):
            raise Exception("invalid handler: %s" % name)
        setattr(self, name, RpcServerAdapter(adaptee=handler, nodeKernel=self.nodeKernel_))

    def _listMethods(self):
        ret = []
        for attrName in dir(self):
            attr = getattr(self, attrName)
            if isinstance(attr, RpcServerAdapter): 
                for methodName in attr._listPublicMethod():
                    ret.append("%s.%s" % (attrName, methodName))
        return ret

    def _methodHelp(self, method):
        for attrName in dir(self):
            attr = getattr(self, attrName)
            if isinstance(attr, RpcServerAdapter): 
                for methodName in attr._listPublicMethod():
                    n = "%s.%s" % (attrName, methodName)
                    if method == n:
                        realMethodProxy = getattr(attr, methodName)
                        assert callable(realMethodProxy)
                        realMethod = realMethodProxy.adaptee_
                        assert callable(realMethod)
                        func = realMethod.im_func
                        defaults = func.func_defaults
                        codeObject = func.func_code
                        argNames = list(codeObject.co_varnames)
                        argCount = codeObject.co_argcount
                        if argNames[0] == "self":
                            argNames = argNames[1:]
                        # add the implicit rpcHeader
                        argNames.insert(0, "rpcHeader")
                        argCount = len(argNames)
                        import pydoc
                        return "%s: doc '%s' argCount %s, argNames %s, defaults %s" % (method,
                                                                                       pydoc.getdoc(realMethod),
                                                                                       argCount, argNames, defaults)

                        
        raise "no such method %s" % method
                              

class NodeXMLRPCRequestHandler(SimpleXMLRPCServer.SimpleXMLRPCRequestHandler):
    def do_POST(self):
        self.server.client_address = self.client_address
        return SimpleXMLRPCServer.SimpleXMLRPCRequestHandler.do_POST(self)

class NodeXMLRPCServer(SimpleXMLRPCServer.SimpleXMLRPCServer):
    def __init__(self, addr):
        self.allow_reuse_address = True
        SimpleXMLRPCServer.SimpleXMLRPCServer.__init__(self, requestHandler=NodeXMLRPCRequestHandler, logRequests=0, addr=addr)

    def _dispatch(self, method, params):
        try:
            host, port = self.client_address[:2]
            if not method.startswith("system"):
                rpcHeader = params[0]
                assertRpcHeaderMightBeValid(rpcHeader=rpcHeader)
                rpcHeader["sourceNetHost"] = host
            return SimpleXMLRPCServer.SimpleXMLRPCServer._dispatch(self, method, params)
        except:
            dsunittest.traceException("converting to xmlrpclib.Fault")
            e = sys.exc_info()[1]
            raise RpcProtocolError(str(e))

    def system_methodSignature(self, method):
        try:
            # check to see if a matching function has been registered
            func = self.funcs[method]
        except KeyError:
            if self.instance is not None:
                # check for a _dispatch method
                # not supported yet
                #                 if hasattr(self.instance, '_dispatch'):
                #                     return "_dispatch signatures not supported"
                #                 else:
                    assert not hasattr(self.instance, '_dispatch')
                    try:
                        func = SimpleXMLRPCServer.resolve_dotted_attribute(
                            self.instance,
                            method
                            )
                    except AttributeError:
                        raise "no such method: %s" % method

        isMemberFunc = False
        assert isinstance(func, RpcServerAdapter) 
        if isinstance(func, RpcServerAdapter): 
            func = func.adaptee_.im_func
            isMemberFunc = True
#         else:
#              try:
#                  func = func.im_func
#                  isMemberFunc = True
#              except AttributeError:
#                  isMemberFunc = False
        defaults = func.func_defaults
        codeObject = func.func_code
        argNames = list(codeObject.co_varnames)
        argCount = codeObject.co_argcount
        if isMemberFunc:
            assert argNames[0] == "self"
            argNames = argNames[1:]
            argCount -= 1

        # add the implicit rpcHeader
        argNames.insert(0, "rpcHeader")

        returnType = "unknown"
        signature = [returnType, "struct"] + (["unknown"] * argCount)


        #return "func is %s argCount %s, argNames %s, defaults %s" % (func, argCount, argNames, defaults)
        return signature
        
class RpcClientAdapter:
    def __init__(self, serverProxy, rpcHeaderAdditions):
        self.serverProxy_ = serverProxy
        self.rpcHeaderAdditions_ = rpcHeaderAdditions
        
    def __getattr__(self, name):
#        try:
        gottenAttr = getattr(self.serverProxy_, name)
#         except AttributeError, e:
#             # this only happens in the case that a node is attempting to Rpc
#             # to itself. Convert to a NetworkError to unify the network and
#             # non-network cases
#             raise xmlrpclib.Fault(str(e), 0)
        return RpcClientAdapter(serverProxy=gottenAttr,
                                rpcHeaderAdditions=self.rpcHeaderAdditions_)
    def __call__(self, *args):
        rpcHeader = args[0]
        rpcHeader = rpcHeader.update(self.rpcHeaderAdditions_)
        args = ConvertAddressToStringRecursively(args)
        try:
            ret = self.serverProxy_(*args)
        except xmlrpclib.Fault, f:
            f.faultString = "%s (%s)" % (rpcHeader, f.faultString)
            raise
        return ConvertStringToAddressRecursively(data=ret)
        
def RpcServerProxyFactory(netAddress, port):
    uri = "http://%s" % netAddress
    serverProxy = xmlrpclib.ServerProxy(uri,transport=xmlrpclib_transport.HTTPTransport(uri))
    return RpcClientAdapter(
        serverProxy=serverProxy, rpcHeaderAdditions={"sourceNetPort": port})

class NodeServer:
    MinPort = 5005
    MaxPort = 5015

    def __init__(self, nodeKernel, ports):
        self.frontend_ = nodeKernel._frontend()

        if ports is None:
            portsToTry = range(self.MinPort, self.MaxPort)
        else:
            portsToTry = ports

        for portI in portsToTry:
            try:
                self.xmlrpcServer_ = NodeXMLRPCServer(addr=('',portI))
            except socket.error, e:
                lastError = e
            else:
                break
        else:
            dsunittest.traceException("could not find a port in portsToTry %s" % portsToTry)
            raise e
        self.port_ = portI
        
        self.xmlrpcServer_.register_introspection_functions()
        self.xmlrpcServer_.register_multicall_functions()
        self.xmlrpcServer_.register_instance(self.frontend_)

    def port(self):
        return self.port_
    
    def serveOneRequest(self):
        self.xmlrpcServer_.handle_request()
    
#defaultBootstrapNetAddress = "root.kenosisp2p.org:5005"
#defaultBootstrapNetAddress = "192.168.0.4:5005"
defaultBootstrapNetAddress = "67.164.12.28:5005"

# Node addresses in Node's public interface are always strings.
class Node(ui.GenericUi):
    def __init__(
        self,
        stringAddress=None, configPath=None,
        ports=None,
        serve=True, stopEvent=None,
        bootstrapNetAddress=defaultBootstrapNetAddress):

        """creates a Kenosis Node. all arguments are optional.

        - do not pass both of stringAddress and configPath.

        - configPath need not exist; it will be created as necessary.

        - ports: an array or tuple of ports to try to bind to. The
        first one that can be bound to will be used.

        - serve: if true the new node will automatically respond to
        requests from other nodes. The only reason to turn this off is
        allow unit testing.

        - stopEvent: changes automatic serving (if serve is
        True) or the serving started by threadedServeUntilEvent (if
        the provided event is None) to use this event. This is useful
        if you want to be able to stop the service later.

        - bootstrapNetAddress: the net address and port (as a string) of
        the node to bootstrap form. Pass None to avoid bootstrapping.

        """

        ui.GenericUi.__init__(self)

        assert not (configPath and stringAddress)

        self.configPath_ = configPath
        if configPath and os.path.exists(configPath):
            stateDict = dsfile.fileObject(path=configPath)
            stringAddress = stateDict["stringAddress"]
            routingNodeNetTuples = stateDict["routingNodeNetTuples"]
        else:
            routingNodeNetTuples = []
            
        if stringAddress is None:
            realNodeAddress = address.randomAddress()
        else:
            assert address.isTextAddress(string=stringAddress), "invalid nodeAddress (string form): %s" % stringAddress
            realNodeAddress = address.NodeAddress(stringAddress=stringAddress)
            
        self.nodeKernel_ = NodeKernel(nodeAddress=realNodeAddress, serverProxyFactory=self.__newServerProxy)
        self.server_ = NodeServer(nodeKernel=self.nodeKernel_, ports=ports)
        self.port_ = self.server_.port()
        if stopEvent is not None:
            self.stopEvent_ = stopEvent
        else:
            self.stopEvent_ = threading.Event()

        for nodeAddress, netAddress in routingNodeNetTuples:
            self.nodeKernel_._updateRoutingTableWith(nodeAddress=nodeAddress, netAddress=netAddress)
        
        if serve:
            self.threadedServeUntilEvent(event=self.stopEvent_)

        if bootstrapNetAddress is not None:
            while not self.stopEvent_.isSet():
                try:
                    self.nodeKernel_._trace("attempting to boostrap from address %s" % bootstrapNetAddress)
                    result = self.bootstrap(netAddress=bootstrapNetAddress)
                except NetworkErrorClasses:
                    dsunittest.traceException("could not bootstrap from address %s" % bootstrapNetAddress)
                    time.sleep(1)
                else:
                    self.nodeKernel_._trace(
                        "succesfully boostrapped from address %s (result %s)" % (bootstrapNetAddress, result))
                    break

    def save(self, configPath=None):
        #"""save to the specified configPath, using the configPath passed to __init__ if None is passed"""
        if configPath is None:
            assert self.configPath_ is not None
            configPath = self.configPath_
            
        routingNodeNetTuples = [
            (n.nodeAddress(), n.netAddress()) for a,b,n in self.nodeKernel_._nodeInfosInBucketAndSuccessors(bucketIndex=0)]
        stateDict = { "stringAddress": self.stringAddress(),
                      "version": 1,
                      "routingNodeNetTuples": routingNodeNetTuples }

        dsfile.setFileObject(path=configPath, object=stateDict)
        
    def __newServerProxy(self, netAddress):
        return RpcServerProxyFactory(netAddress=netAddress, port=self.port_)

    # setup methods
    def registerNamedHandler(self, name, handler):
        self.server_.frontend_.registerNamedHandler(name=name, handler=handler)

    def bootstrap(self, netAddress):
        return self.nodeKernel_.bootstrap(netAddress=netAddress)

    def stepUntilEvent(self, event):
        while not event.isSet():
            self.step()
            event.wait(1.0)

    def step(self):
        self.nodeKernel_.step()

    def serveOneRequest(self):
        return self.server_.serveOneRequest()

    def serveUntilEvent(self, event):
        while not event.isSet():
            self.serveOneRequest()

    def threadedServeUntilEvent(self, event=None):
        if event is None:
            event = self.stopEvent_
        self.thread_ = dsthread.newThread(function=self.serveUntilEvent, params=(event,))
        self.stepThread_ = dsthread.newThread(function=self.stepUntilEvent, params=(event,))
        
    def rpc(self, stringAddress=None):
        nodeAddress = address.NodeAddress(stringAddress=stringAddress)
        return self.nodeKernel_.rpc(nodeAddress=nodeAddress)

    def port(self):
        return self.server_.port()

    def findNearestNodes(self, stringAddress, serverProxyFilterAndMapFunc=None):
        """Returns an array of pairs of node address string and the
        network address and port of the node with that address."""
        nodeAddress = address.NodeAddress(stringAddress=stringAddress)
        nodeAddressNetAddressPairs = self.nodeKernel_._findNearestNodeAddressNetAddressTuples(
            nodeAddress=nodeAddress, serverProxyFilterAndMapFunc=serverProxyFilterAndMapFunc)
        return [(nodeAddress.numericRepr(), netAddress) for
                nodeAddress, netAddress in nodeAddressNetAddressPairs]

    def stringAddress(self):
        return self.nodeKernel_.nodeAddress_.numericRepr()
    
class NodeKernel:
    def __init__(self, nodeAddress, serverProxyFactory):
        self.nodeAddress_ = nodeAddress
        self.buckets_ = []
        for i in range(address.addressLengthInBits):
            self.buckets_.append([])

        self.bucketLock_ = threading.RLock()
        self.constantsK_ = 20
        self.constantsAlpha_ = 3
        self.staleInfoTime_ = 30
        self.needPingPairs_ = []
        self.serverProxyFactory_ = serverProxyFactory
        self.frontend_ = NodeRpcFrontend(nodeKernel=self)
        self.taskList_ = task.TaskList(maxThreads=2 * self.constantsAlpha_)
        self.taskList_.start(wait=0)
        self.newResultsEvent_ = dsthread.MultithreadEvent()
        self.results_ = {}

    def __repr__(self):
        return "NodeKernel(%s)" % self.nodeAddress_
    
    def rpc(self, nodeAddress):
        rpcHeader=self._rpcHeaderFor(nodeAddress=nodeAddress)
        if nodeAddress == self.nodeAddress_:
            serverProxy=self.frontend_
            rpcHeader["sourceNetHost"] = "localhost"
            rpcHeader["sourceNetPort"] = 123
        else:
            serverProxy=self._serverProxyForNodeAddress(nodeAddress=nodeAddress)
            
        return Rpc(
            rpcHeader=rpcHeader,
            serverProxy=serverProxy)

    def rpcFindNode(self, nodeAddress):
        dsunittest.assertNotEqual(nodeAddress, self.nodeAddress_)
        bucketI = self._bucketIndexForNodeAddress(nodeAddress=nodeAddress)
        ret = []
        for a, b, ni in self._nodeInfosInBucketAndSuccessors(bucketIndex=bucketI):
            ret.append((ni.nodeAddress(), ni.netAddress()))
            if len(ret) >= self.constantsK_:
                break
        return ret

    def _rpcHeaderFor(self, nodeAddress):
        return newRpcHeader(sourceNodeAddress=self.nodeAddress_, destNodeAddress=nodeAddress)

    def _bucketIndexForNodeAddress(self, nodeAddress):
        assert nodeAddress != self.nodeAddress_
        assert nodeAddress != NodeAddressUnknown
        distance = address.distance(address0=self.nodeAddress_, address1=nodeAddress)
        for i in range(address.addressLengthInBits):
            if 2**i <= distance and distance < 2**(i+1):
                return i
        raise address.InvalidAddressError("distance is surprisingly large: %s" % (distance))

    # return constantsK_ nearest node infos
    def _findNearestNodeAddressNetAddressTuples(self, nodeAddress, serverProxyFilterAndMapFunc=None):
        self._trace(text="_findNearestNodeAddressNetAddressTuples >>: %s" % nodeAddress, prio=1)
        try:
            ret = self.__findNodeNetAddress(nodeAddress=nodeAddress, requireExactMatch=False,
                                            serverProxyFilterAndMapFunc=serverProxyFilterAndMapFunc)
            def sorter(a,b,nodeAddress=nodeAddress):
                distA = address.distance(a[0], nodeAddress)
                distB = address.distance(b[0], nodeAddress)
                r = cmp(distA, distB)
                return r

            ret.sort(sorter)
        finally:
            self._trace(text="_findNearestNodeAddressNetAddressTuples <<: %s" % nodeAddress, prio=1)
        return ret
        
    def _findNodeNetAddress(self, nodeAddress):
        self._trace(text="_findNodeNetAddress >>: %s" % nodeAddress, prio=1)
        ret = None
        try:
            ret = self.__findNodeNetAddress(nodeAddress=nodeAddress)
        finally:
            self._trace(text="_findNodeNetAddress <<: %s" % ret, prio=1)
        return ret

    def __findNodeNetAddress(self, nodeAddress, requireExactMatch=True, serverProxyFilterAndMapFunc=None):

        """If requireExactMatch is true returns the net address of
        the node with nodeAddress or raises NodeNotFound.

        If requireExactMatch is not true returns an array of
        (nodeAddress, netAddress) tuples in order of increasing
        distance from the requested nodeAddress."""

        if nodeAddress == self.nodeAddress_:
            bucketI = 0
        else:
            bucketI = self._bucketIndexForNodeAddress(nodeAddress=nodeAddress)
        currentSet = Set()
        successDict = {}
        successNodeInfo = None
        errorSet = Set()
        realQueue = dsqueue.PriorityQueue(0)
        mergeQueue = GrowingQueue(
            realQueue=realQueue, iterator=self._nodeInfosInBucketAndSuccessors(bucketI))
        while len(successDict) < self.constantsK_:
            while len(currentSet) < self.constantsAlpha_ and not mergeQueue.empty():
                bucketIndex, freshness, nodeInfo = mergeQueue.get()
                self._trace("mergeQueue get: %s" % repr(nodeInfo))
                currentNodeAddress = nodeInfo.nodeAddress()
                currentNetAddress = nodeInfo.netAddress()
                del nodeInfo
                if currentNodeAddress in successDict:
                    self._trace("  current node is in success dict")
                    continue
                if (currentNodeAddress, currentNetAddress) in (currentSet | errorSet):
                    self._trace("  current node, net address is in current, error dict")
                    continue

                currentTuple = (currentNodeAddress, currentNetAddress)
                currentSet.add(currentTuple)
                
            if len(currentSet) == 0 and mergeQueue.empty():
                self._trace("break because currentSet is empty and mergeQueue is empty")
                break

            self.newResultsEvent_.clear()
            anyCommandsCompleted = False
            for calledTuple in tuple(currentSet):
                calledNodeAddress, calledNetAddress = calledTuple
                try:
                    if calledNodeAddress == nodeAddress:
                        result = self._nodeCommandResult(
                            netAddress=calledNetAddress, nodeAddress=calledNodeAddress,
                            commandName="ping")
                    else:
                        result = self._nodeCommandResult(
                            netAddress=calledNetAddress, nodeAddress=calledNodeAddress,
                            commandName="findNode", commandArgs=(nodeAddress,))
                except _StaleInfo:
                    continue
                except NetworkErrorClasses:
                    dsunittest.traceException(text="Error while talking to node");
                    result = None
                anyCommandsCompleted = True
                currentSet.remove((calledNodeAddress, calledNetAddress))
                self._trace("called %s, foundNodes=%s (desired=%s)" % (repr(calledTuple),
                                                                            result,
                                                                            nodeAddress))
                if result is not None:
                    self._updateRoutingTableWith(
                        nodeAddress=calledNodeAddress, netAddress=calledNetAddress)
                    if calledNodeAddress == nodeAddress:
                        if requireExactMatch:
                            # we called and heard back from the right node!
                            return calledNetAddress
                    else:
                        for foundNodeAddress, netAddress in result:
                            if foundNodeAddress == self.nodeAddress_:
                                continue
                            resultNodeInfo = NodeInfo(
                                nodeAddress=foundNodeAddress, netAddress=netAddress)
                            resultTuple = (0, 0, resultNodeInfo)
                            self._trace("adding resultTuple %s to realQueue %s" % (resultTuple, realQueue))
                            realQueue.put(resultTuple)
                    successDict[calledNodeAddress] = calledNetAddress
                    if len(successDict) == self.constantsK_:
                        break
                else:
                    errorSet.add((calledNodeAddress, calledNetAddress))

            # If no commands have completed then there is nothing to
            # do until one does complete.
            if not anyCommandsCompleted:
                self._trace("Waiting on newResultsEvent_")
                self.newResultsEvent_.wait()
                self._trace("Finished waiting on newResultsEvent_")
        assert len(currentSet) == 0 or len(successDict) == self.constantsK_
        if requireExactMatch:
            assert not nodeAddress in successDict
            raise NodeNotFound(nodeAddress)
        else:
            return successDict.items()

    def __threadsafeCopyOfBucket(self, bucketIndex):
        self.bucketLock_.acquire()
        try:
            return copy.copy(self.buckets_[bucketIndex])
        finally:
            self.bucketLock_.release()
            
    def _nodeInfosInBucketAndSuccessors(self, bucketIndex):
        for i in range(bucketIndex, len(self.buckets_)):
            for nodeInfo in self.__threadsafeCopyOfBucket(bucketIndex=i):
                yield (i, nodeInfo.freshness(), nodeInfo)
        for i in range(bucketIndex-1, 0-1, -1):
            for nodeInfo in self.__threadsafeCopyOfBucket(bucketIndex=i):
                yield (i, nodeInfo.freshness(), nodeInfo)
        
    def _nextHopNetAddressForNodeAddress(self, nodeAddress):
        ni = self._nodeInfoForNodeAddress(nodeAddress=nodeAddress)
        if ni:
            return ni.netAddress()
                                          
        netAddress = self._findNodeNetAddress(nodeAddress=nodeAddress)
        return netAddress

    def _serverProxyForNodeAddress(self, nodeAddress):
        netAddress = self._nextHopNetAddressForNodeAddress(nodeAddress=nodeAddress)
        return self._serverProxyForNetAddress(netAddress=netAddress)

    def _serverProxyForNetAddress(self, netAddress):
        return self.serverProxyFactory_(netAddress=netAddress)

    def _updateRoutingTableWith(self, nodeAddress, netAddress):
        self.bucketLock_.acquire()
        try:
            bucketIndex = self._bucketIndexForNodeAddress(nodeAddress)
            bucket = self.buckets_[bucketIndex]

            for nodeInfo in bucket:
                if nodeInfo.nodeAddress() == nodeAddress:
                    i = bucket.index(nodeInfo)
                    n2 = bucket[i]
                    n2.updateFreshness(newNetAddress=netAddress)
                    break
            else:
                nodeInfo = NodeInfo(nodeAddress=nodeAddress, netAddress=netAddress)
                if len(bucket) < self.constantsK_:
                    bucket.append(nodeInfo)
                else:
                    toRemoveNodeInfo = bucket[-1]
                    toInsertTuple = (nodeInfo, bucket)
                    self.needPingPairs_.append(toInsertTuple)
                    return
            self._sortBucket(bucket=bucket)
        finally:
            self.bucketLock_.release()


    def _considerNeedPingPairs(self):
        for newNi, bucket in self.needPingPairs_:
            oldNi = bucket[-1]
            try:
                result = self._pingNode(nodeAddress=oldNi.nodeAddress(), netAddress=oldNi.netAddress())
                assert result
            except NetworkErrorClasses:
                self.bucketLock_.acquire()
                try:
                    bucket.remove(oldNi)
                    bucket.append(newNi)
                finally:
                    self.bucketLock_.release()
            else:
                oldNi.updateFreshness()
            self._sortBucket(bucket=bucket)
        self.needPingPairs_ = []

    def _sortBucket(self, bucket):
        self.bucketLock_.acquire()
        try:
            bucket.sort(lambda x,y:cmp(x.freshness(), y.freshness()))
        finally:
            self.bucketLock_.release()

    def step(self):
        self._considerNeedPingPairs()
                
    def _nodeInfoForNodeAddress(self, nodeAddress):
        bucketIndex = self._bucketIndexForNodeAddress(nodeAddress)
        self.bucketLock_.acquire()
        try:
            for ni in self.buckets_[bucketIndex]:
                if ni.nodeAddress() == nodeAddress:
                    return ni
            return None
        finally:
            self.bucketLock_.release()

    def _pingNode(self, nodeAddress, netAddress):
        self._trace(text="_pingNode >>: %s" % nodeAddress, prio=1)
        sp = self._serverProxyForNetAddress(netAddress=netAddress)
        na = sp.kenosis.ping(self._rpcHeaderFor(nodeAddress=nodeAddress))
        assert na == nodeAddress, "%s != %s" % (na, nodeAddress)
        self._trace(text="_pingNode <<: %s" % nodeAddress, prio=1)
        return True

    def _callRemoteFindNode(self, nextHopNodeAddress, nextHopNetAddress, nodeAddress):
        self._trace(text="_callRemoteFindNode >>: %s" % (repr(locals())), prio=1)
        sp = self._serverProxyForNetAddress(netAddress=nextHopNetAddress)
        ret = sp.kenosis.findNode(
            self._rpcHeaderFor(nodeAddress=nextHopNodeAddress), nodeAddress)
        self._trace(text="_callRemoteFindNode <<", prio=1)
        return ret

    def bootstrap(self, netAddress):
        try:
            self._trace(text="bootstrap >>: %s" % netAddress, prio=1)
            sp = self._serverProxyForNetAddress(netAddress=netAddress)
            nodeAddress = sp.kenosis.ping(self._rpcHeaderFor(nodeAddress=NodeAddressUnknown))
            if nodeAddress == self.nodeAddress_:
                self._trace("NodeKernel.bootstrap returning false because pinged node has my address")
                return False
            else:
                self._updateRoutingTableWith(netAddress=netAddress, nodeAddress=nodeAddress)
                return True
        finally:
            self._trace(text="bootstrap <<: %s" % netAddress, prio=1)

    def _frontend(self):
        return self.frontend_


    def _nodeCommandResult(self, nodeAddress, netAddress, commandName, commandArgs=()):
        self._trace("_nodeCommandResult(%s)" % locals())
        key = (nodeAddress, netAddress, commandName, commandArgs)
        try:
            resultInfo = self.results_[key]
        except KeyError:
            pass
        else:
            if dstime.time() - resultInfo["time"] < self.staleInfoTime_:
                try:
                    return resultInfo["result"]
                except KeyError:
                    raise resultInfo["error"]
            else:
                self._trace("resultInfo %s is stale" % resultInfo)
        self._addWork(
            nodeAddress=nodeAddress, netAddress=netAddress,
            commandName=commandName, commandArgs=commandArgs)
        raise _StaleInfo()

    def _setNodeCommandResult(self, nodeAddress, netAddress, commandName, commandArgs,
                              result=None, error=None):
        key = (nodeAddress, netAddress, commandName, commandArgs)
        resultInfo = {"time":dstime.time()}
        if error:
            resultInfo["error"] = error
        else:
            resultInfo["result"] = result
        self.results_[key] = resultInfo
        self._trace("set resultInfo %s for key %s" % (resultInfo,key))
        self.newResultsEvent_.set()

    def _addWork(self, nodeAddress, netAddress, commandName, commandArgs=()):
        self._trace("> addWork(): %s" % ((nodeAddress, netAddress, commandName, commandArgs),))
        if commandName == "ping":
            work = self._pingNode
        elif commandName == "findNode":
            work = self._callRemoteFindNode
        else:
            # For some reason this exception cannot be caught when it
            # has a space in it, though other strings can be.
            raise "invalidcommandName"

        def lf(work=work, args=commandArgs, nodeAddress=nodeAddress, netAddress=netAddress):
            try:
                result = work(nodeAddress, netAddress, *args)
                error = None
            except NetworkErrorClasses, e:
                error = e
                result = None
            self._setNodeCommandResult(
                nodeAddress=nodeAddress, netAddress=netAddress,
                commandName=commandName,
                commandArgs=commandArgs,
                result=result,
                error=error)
                
        self.taskList_.addCallableTask(lf, id=(nodeAddress, netAddress, commandName, commandArgs))

    def _trace(self, text, prio=2):
        dsunittest.trace(text="%s: %s" % (self, text), prio=prio)

